#!/usr/bin/perl -w

use strict;
use File::Copy;
use POSIX qw();
use Data::Dumper;
use URI;
use File::Basename;
use Net::FTP;
use Net::SCP::Expect;
use LWP::UserAgent;
use Unicode::MapUTF8 qw(to_utf8 from_utf8);
use MIME::Lite;
use Net::SMTP;
use Net::Domain;
use Config::Properties;

##******************************************************************************
## Variables globales de configuration
##******************************************************************************

my $rootInstall = "/tmp/export/";
$rootInstall = "/var/lib/interldap/atransport";

# Repertoire ou prendre les fichiers :
my $inputDir  = "$rootInstall/input/";   # mettre un / a la fin !
# Repertoire des files d'attente :
my $fifosDir  = "$rootInstall/fifos/";   # mettre un / a la fin !
# Repertoire ou mettre les fichiers transmis :
my $doneDir   = "$rootInstall/done/";    # mettre un / a la fin !
# Repertoire ou mettre les fichiers en erreur :
my $errorDir  = "$rootInstall/error/";   # mettre un / a la fin !

# Nombre de secondes entre chaque parcours du repertoire $inputDir
my $timeBetweenDistributions = 2;

# Nombre de secondes entre envoi dans une file donnee :
my $timeBetweenSending = 1;

# Mettre l'option -d pour avoir des messages de debugage :
my $debug = 0;
$debug = 1  if( exists( $ARGV[0] ) and $ARGV[0] eq '-d' );

# Fichier de proprietes
my $confDirectory = "/etc/interldap";
my $confFilename = "$confDirectory/atransport-engine.properties";

##******************************************************************************
## Utilisation du fichier de proprietes
##******************************************************************************
open( PROPS, "<$confFilename" ) or die("open $confFilename : $!");
my $properties = new Config::Properties();
$properties->load(*PROPS);
my $treatment = $properties->getProperty( "export.treatment", '/bin/true' );
my $mailFromName = $properties->getProperty( "export.mail_from_name", 'Moteur d\'exportation' );
my $mailFromAdress = $properties->getProperty( "export.mail_from_adress", 'root@localhost' );
my $mailSubject = $properties->getProperty( "export.mail_subject", 'Transfert d\'un fichier' );
close(PROPS);
$properties = undef;

my $user = $ENV{USER};
my $host = Net::Domain::hostfqdn();
$mailFromAdress =~ s/%u/$user/g;
$mailFromAdress =~ s/%h/$host/g;
($user,$host) = ();

warn("treatment : $treatment\n") if( $debug );
warn("mailFromName : $mailFromName\n") if( $debug );
warn("mailFromAdress : $mailFromAdress\n") if( $debug );
warn("mailSubject : $mailSubject\n") if( $debug );

##******************************************************************************
## Autres variables globales
##******************************************************************************
my $tmpDir    = undef;

my %serverPid = ();
my @killedPid = ();

##******************************************************************************
## Fonction getInfosFromURI (leve des exceptions)
## extrait les informations de l'URI
##******************************************************************************
sub getInfosFromURI
{
   my ($destinationString) = @_;
   my ($scheme, $user, $passwd, $host, $path) =
         ( $destinationString =~ m|^([^:]+)://([^:@/]+)(?::([^:@/]+))?@([^@/]*)(.*)$| );
   if( !defined( $path ) )
   {
      my ($server, $port, $email) =
            ( $destinationString =~ m|^mail://(?:([^:/]+)(?::([^:@/]+))?)?/(.*)$| );
      die("Bad format for URI : $destinationString") if( ! defined( $email ) );
      return { scheme=>'mail', server=>$server, port=>$port, email=>$email };
   }
   return { scheme=>$scheme, user=>$user, passwd=>$passwd, host=>$host, path=>$path };
}

##******************************************************************************
## Fonction sendHTTP (leve des exceptions)
## envoyer via HTTP
##******************************************************************************
sub sendHTTP
{
   my ($destinationInfos, $baseName, $tmpFileName) = @_;

   my $url = 'http://'.$destinationInfos->{user}.':'.$destinationInfos->{passwd}.'@'.$destinationInfos->{host}.$destinationInfos->{path};

   my $ua = LWP::UserAgent->new();
   my $res=$ua->post("$url",
       Content_Type => 'multipart/form-data',
       Content      => [ user       => $destinationInfos->{user},
                         password   => $destinationInfos->{passwd},
                         nomfichier => $baseName,
                         fichier    => [$tmpFileName],
                        ],
       ) or die("ua->post(): $!\n");

   die( $res->error_as_HTML() )
      if( ! $res->is_success() );
}

##******************************************************************************
## Fonction sendSCP (leve des exceptions)
## envoyer via SCP
##******************************************************************************
sub sendSCP
{
   my ($destinationInfos, $baseName, $tmpFileName) = @_;

   my $pid = fork();
   die("$0: fork: $!") if( $pid == -1 );
   if( $pid )
   {  # Pere
      wait();
   }
   else
   {  # Fils
      open( STDIN, "</dev/null" );
      open( STDOUT, ">/dev/null" );
      open( STDERR, ">/dev/null" );
      {
         exec( "/usr/bin/scp",
               $tmpFileName,
               "$destinationInfos->{user}\@$destinationInfos->{host}:$destinationInfos->{path}" );
      }
      exit( 1 );
   }
}


#sub sendSCP
#{
#   my ($destinationInfos, $baseName, $tmpFileName) = @_;
#
#   my $scpe = Net::SCP::Expect->new( host     => $destinationInfos->{host},
#                                     user     => $destinationInfos->{user},
#                                     password => $destinationInfos->{passwd} )
#      or die( "Net::SCP::Expect->new(): $!" );
#
#   $scpe->scp( $tmpFileName, $destinationInfos->{path} )
#      or die("scpe->scp(): $!");
#}

##******************************************************************************
## Fonction sendFTP (leve des exceptions)
## envoyer via FTP
##******************************************************************************
sub sendFTP
{
   my ($destinationInfos, $baseName, $tmpFileName) = @_;

   my $ftp = Net::FTP->new( $destinationInfos->{host}, Debug => 0, Passive => 1 )
      or die "Net::FTP->new(): $@";

   $ftp->login( $destinationInfos->{user}, $destinationInfos->{passwd} )
      or die("ftp->login(): ". $ftp->message());

   $ftp->cwd( $destinationInfos->{path} )
      or die("ftp->cwd(): ",$ftp->message());

   $ftp->binary()
      or die("ftp->binary(): ", $ftp->message());

   $ftp->put( $tmpFileName, $baseName )
      or die("ftp->put(): ", $ftp->message());

   $ftp->quit()
      or die("ftp->quit(): ", $ftp->message());
}

##******************************************************************************
## Fonction sendSMTP (l�ve des exceptions)
## envoyer via SMTP
##******************************************************************************
sub sendSMTP
{
   my ($destinationInfos, $baseName, $tmpFileName) = @_;

   # Valeurs par d�faut :
   $destinationInfos->{server} ||= 'localhost';
   $destinationInfos->{port}   ||= '25';

   my $subject = $mailSubject;
   $subject =~ s/%f/$baseName/g;
   warn("subject : $subject\n") if( $debug );

   my $mime = MIME::Lite->new(
      From    => $mailFromName.'<'.$mailFromAdress.'>',
      To      => $destinationInfos->{email},
      Subject => $subject,
      Type    => 'multipart/mixed',
   );
   $mime->attach(
      Type       => 'application/xml',
      Encoding   => 'base64',
      Path       => $tmpFileName,
      Filename   => $baseName,
   );

   my $smtp = Net::SMTP->new( $destinationInfos->{server},
                              Port => $destinationInfos->{port} )
      or die("Net::SMTP->new(): $!");

   $smtp->mail($mailFromAdress)          or die("smtp->mail(): $!");
   $smtp->to($destinationInfos->{email}) or die("smtp->to(): $!");
   $smtp->data()                         or die("smtp->data(): $!");
   $smtp->datasend( $mime->as_string() ) or die("smtp->datasend(): $!");
   $smtp->dataend()                      or die("smtp->dataend(): $!");
   $smtp->quit()                         or die("smtp->quit(): $!");

}

##******************************************************************************
## Variable senderFunction
## repartiteur selon le protocole
##******************************************************************************
my $senderFunction =
{
   ftp  => \&sendFTP,
   http => \&sendHTTP,
   scp  => \&sendSCP,
   sftp => \&sendSCP,
   mail => \&sendSMTP,
};

##******************************************************************************
## Fonction extractDate
## utile pour trier les fichiers selon le numero de secondes qu'ils ont dans le nom
##******************************************************************************
sub extractDate
{
   my ($file) = @_;
   my ($export,$date) = ( $file =~ m:^.*/([\w-]+?)(\d+)[^/]*$: );
   return int($date);
}

##******************************************************************************
## Fonction senderMain
## fonction principale d'un sender
##******************************************************************************
sub senderMain
{
   my ($export) = @_;
   $0 .= " sender for $export"; # pour le voir dans ps

   my %waitingFiles = ();

   while(1)
   {
      LABELFICHIER: foreach my $file (sort {extractDate($a)<=>extractDate($b)} glob($fifosDir.$export.'/*.hdr'))
      {
         my $baseName;
         my $tmpFileName;
         my $dataFile = $file; 

         eval # Gestion des exceptions
         {

            # --------------------------------------------------------
            # Lecture du fichier d'en-tête
            open(FILE, $file) or die("open($file) : $!");

            my $ligne = <FILE>;
            my ($extension) = ( $ligne =~ m/^Extension\s*:\s*(.+)$/ );
            defined($extension) or die("Extension line error in $file");
            $dataFile =~ s/hdr$/$extension/;
            $baseName = basename($dataFile);

            # Récupération de l'encodage
            $ligne = <FILE>;
            my ($encodingString) = ( $ligne =~ m/^Encoding\s*:\s*(.+)$/ );
            defined($encodingString) or die("Encoding line error in $file");
            my ($carriageReturn,$encoding) = ( $encodingString =~ m/^\{(.)\}(.*)$/ );

            # Récupération de la destination
            $ligne = <FILE>;
            my ($destinationString) = ( $ligne =~ m/^URI\s*:\s*(.+)$/ );

            # Si pas de destination : c'est que c'était la dernière : done
            if( ! defined($destinationString) )
            {
               close FILE;
               move( $file, $doneDir );
               move( $dataFile, $doneDir );
            }
            else
            {

               # --------------------------------------------------------
               # Modification du fichier
               open(FILE2, ">$file.tmp") or die("open($file) : $!");
               print FILE2 "Extension: $extension\n";
               print FILE2 "Encoding: $encodingString\n";
               while( $ligne = <FILE> )
               {
                  print FILE2 $ligne;
               }
               close FILE;
               close FILE2;
               rename( "$file.tmp", $file ) or die("rename( $file.tmp, $file ): $!");

               # --------------------------------------------------------
               # Traitement de l'encodage et du système
               open(FILE, $dataFile) or die("open($file) : $!");
               my $content;
               {
                  local $/ = undef;
                  $content = <FILE>;
               }
               close(FILE);

               $content = from_utf8({-string=>$content,-charset=>'ISO-8859-1'}); # Obligatoire pour
               $content = to_utf8  ({-string=>$content,-charset=>'ISO-8859-1'}); # avoir de l'UTF
               if( $encoding !~ /^UTF/ )
               {
                  $content = from_utf8({-string=>$content,-charset=>'ISO-8859-1'});
               }
               if( $carriageReturn eq 'W' )
               {
                  $content =~ s/\n/\r\n/g;
               }

               # Copie modifiée du fichier dans $tmpDir
               $tmpFileName = $tmpDir.'/'.$baseName;
               open( TMPFILE, ">$tmpFileName" ) or die("open $tmpFileName: $!");
               binmode TMPFILE;
               print( TMPFILE "$content" );
               close( TMPFILE );

               # --------------------------------------------------------
               # Ex�cution du traitement
               my $execLine = $treatment;
               $execLine =~ s/%f/$tmpFileName/g;
warn("system $execLine\n") if( $debug );
               {
                  local $SIG{CHLD} = 'DEFAULT';
                  system($execLine);
               }

               my $destinationInfos = getInfosFromURI( $destinationString );
               my $scheme = $destinationInfos->{scheme};
               die( "Unknown scheme '$scheme' file $file\n" )
                  if( ! exists( $senderFunction->{$scheme} ) );
               $senderFunction->{$scheme}->( $destinationInfos, $baseName, $tmpFileName );

            }

         }; # Fin du eval de gestion des exceptions

         if( $@ )  # Si il y a une exception
         {
warn("Exception : $file : $@\n") if( $debug );
            if( ! exists( $waitingFiles{$baseName} ) )
            {  # Première erreur :
               $waitingFiles{$baseName} = 1;
            }
            else
            {  # Seconde erreur :
               delete( $waitingFiles{$baseName} );
               move( $file, $errorDir );
               move( $dataFile, $errorDir );
            }
         }

         unlink( $tmpFileName ) if( defined($tmpFileName) );

      }

      sleep( $timeBetweenSending );
   }
}

##******************************************************************************
## Fonction senderCreate
## pour créer un sender sur un export donné
##******************************************************************************
sub senderCreate
{
   my ($export) = @_;

   my $pid = fork();
   die("$0: fork: $!") if( $pid == -1 );
   if( $pid )
   {  # Pere
      $serverPid{$pid} = $export;
   }
   else
   {  # Fils
      $SIG{CHLD} = 'IGNORE';
      senderMain( $export );
      die("$0: should not be here");
   }
}

##******************************************************************************
## Fonction recreateKilledSenders
## recrée les senders qui seraient morts
##******************************************************************************
sub recreateKilledSenders
{
   my @killedPidCopy = @killedPid;
   foreach my $pid (@killedPidCopy)
   {
      my $export = $serverPid{$pid};
      delete $serverPid{$pid};
      senderCreate( $export );
      @killedPid = grep { $_ != $pid } @killedPid;
   }
}

##******************************************************************************
## Fonction fileDistribution
## dispatche les fichiers d'input dans les répertoires de fifos
## appelle senderCreate si nouveau export
##******************************************************************************
sub fileDistribution
{
   my ($inputDir,$fifosDir) = @_;

   # Lecture des fichier :
   opendir(DIR, $inputDir) or die "opendir $inputDir: $!";
   my @inputFiles = grep( m/\.hdr$/, grep {-f} map {$inputDir.$_} readdir(DIR));
   closedir DIR;

   # Traitement des fichiers :
   foreach my $file ( @inputFiles )
   {
      if( my ($export,$date) = ( $file =~ m:^.*/([\w-]+?)(\d+)[^/]*$: ) )
      {
warn "$file $export $date\n" if( $debug );

         if( ! -d $fifosDir.$export )
         {
            if( ! mkdir( $fifosDir.$export ) )
            { warn("$0: mkdir $fifosDir.$export: $!\n") if( $debug ); }
            senderCreate( $export );
         }

         # Recherche du nom de son fichier de données
         my $dataFile = $file;
         if( open(FILE, $file) )
         {
            my $ligne = <FILE>;
            my ($extension) = ( $ligne =~ m/(\S+)$/ );
            $dataFile =~ s/hdr$/$extension/;
            close FILE;
         }
         else
         {
            warn("$0: $file: $!\n") if( $debug );
            move( $file, $errorDir );
         }

         if( ! move($file,$fifosDir.$export) )
         {
            warn("$0: move($file,$fifosDir.$export): $!\n") if( $debug );
         }
         elsif( ! move($dataFile,$fifosDir.$export) )
         {
            warn("$0: move($dataFile,$fifosDir.$export): $!\n") if( $debug );
         }

      }
      else
      {
         warn("$0: $file: Bad name format\n") if( $debug );
         move( $file, $errorDir );
      }
   }
}

##******************************************************************************
## Fonction sigChildHandler
## appelée lors d'un signal SIGCHILD
##******************************************************************************
sub sigChildHandler
{
   $SIG{CHLD} = \&sigChildHandler;

   my $child;
   while( ( $child = waitpid(-1,&POSIX::WNOHANG) ) > 0 )
   {
      push @killedPid, $child;
   }
}
$SIG{CHLD} = \&sigChildHandler;

##******************************************************************************
## main
##******************************************************************************

# Création du répertoire TMP
$tmpDir = POSIX::tmpnam();
mkdir( $tmpDir ) or die("mkdir $tmpDir: $!");

$SIG{INT} = sub{ rmdir( $tmpDir ); exit(0); };

# Création des senders pour les répertoires existants
opendir(DIR, $fifosDir) or die "opendir $fifosDir: $!";
my @exportDir = grep { !/^\./ && -d $fifosDir.$_ } readdir(DIR);
closedir DIR;
map { senderCreate( $_ ) } @exportDir;

# Attente de nouveaux fichier
while(1)
{
   fileDistribution( $inputDir, $fifosDir);
   recreateKilledSenders();
   sleep( $timeBetweenDistributions );
}

##******************************************************************************
